#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
All rights reserved.
Contact: Nokia Corporation (qt-info@nokia.com)

This file is part of the documentation of Qt. It was originally
published as part of Qt Quarterly.

$QT_BEGIN_LICENSE:BSD$
You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
  * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
    the names of its contributors may be used to endorse or promote
    products derived from this software without specific prior written
    permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
$QT_END_LICENSE$
"""

import sys
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.QtXmlPatterns import *

class Books:

    def __init__(self, books):
    
        self.books = books
        for book in books:
            book.parent = self
        
        self.parent = None
    
    def index(self, value):
        return self.books.index(value)

class Book:

    def __init__(self, title, authors):
    
        self.title = Field(u"title", title, self)
        self.authors = Field(u"authors", authors, self)
        self.parent = None
    
    def index(self, value):
    
        if self.title == value:
            return 0
        elif self.authors == value:
            return 1
        else:
            raise ValueError, repr(value) + " not a field in this book"

class Field:

    def __init__(self, name, data, parent):
    
        self.name = name
        self.text = Text(data, self)
        self.parent = parent
    
    def index(self, value):
        if self.text == value:
            return 0
        else:
            raise ValueError, repr(value) + " not a field value"

class Text:

    def __init__(self, data, parent):
    
        self.data = data
        self.parent = parent


class BookModel(QSimpleXmlNodeModel):

    def __init__(self, books, namePool):
    
        QSimpleXmlNodeModel.__init__(self, namePool)
        
        self.books = books
        self.books.parent = self
        self.cache = {}
    
    def documentUri(self, node_index):
    
        item = self._getItem(node_index)
        if item == self:
            return QUrl("books://")
        else:
            return QUrl()
    
    def _createIndex(self, item):
    
        self.cache[id(item)] = item
        return self.createIndex(id(item))
    
    def _getItem(self, node_index):
        pointer = node_index.data()
        if pointer:
            return self.cache.get(pointer)
        else:
            return None
    
    def root(self, node_index = None):
        return self._createIndex(self)
    
    def name(self, node_index):
    
        item = self._getItem(node_index)
        if not item:
            return QXmlName()
        
        if isinstance(item, Books):
            return QXmlName(self.namePool(), "books")
        elif isinstance(item, Book):
            return QXmlName(self.namePool(), "book")
        elif isinstance(item, Field):
            return QXmlName(self.namePool(), item.name)
        else:
            return QXmlName()
    
    def kind(self, node_index):
    
        item = self._getItem(node_index)
        
        if isinstance(item, Text):
            return QXmlNodeModelIndex.Text
        elif item == self:
            return QXmlNodeModelIndex.Document
        else:
            return QXmlNodeModelIndex.Element
    
    def attributes(self, element):
    
        return []
    
    def stringValue(self, node_index):
    
        item = self._getItem(node_index)
        if isinstance(item, Text):
            return item.data
        else:
            return QSimpleXmlNodeModel.stringValue(self, node_index)
    
    def typedValue(self, node_index):
    
        item = self._getItem(node_index)
        if isinstance(item, Text):
            return QVariant(item.data)
        else:
            return QVariant()
    
    def compareOrder(self, ni1, ni2):
    
        if ni1 == ni2:
            return QXmlNodeModelIndex.Is
        
        item1 = self._getItem(ni1)
        item2 = self._getItem(ni2)
        
        # Check for one of the nodes being the document itself.
        if item1 == self:
            return QXmlNodeModelIndex.Precedes
        
        if item2 == self:
            return QXmlNodeModelIndex.Follows
        
        # Check the location of each node within the structure.
        location1 = []
        while item1.parent is not self:
            location1.append(item1.parent.index(item1))
            item1 = item1.parent
        
        location2 = []
        while item2.parent is not self:
            location2.append(item2.parent.index(item2))
            item2 = item2.parent
        
        while location1 and location2:
        
            i1 = location1.pop()
            i2 = location2.pop()
            
            if i1 < i2:
                return QXmlNodeModelIndex.Precedes
            elif i1 > i2:
                return QXmlNodeModelIndex.Follows
        
        return QXmlNodeModelIndex.Is
    
    def nextFromSimpleAxis(self, axis, origin):
    
        item = self._getItem(origin)
        if not item:
            return QXmlNodeModelIndex()
        
        return_item = None
        
        if isinstance(item, Text):
        
            if axis == self.Parent:
                return_item = item.parent
        
        else:
        
            if axis == self.Parent:
                return_item = item.parent
            
            elif axis == self.FirstChild:
            
                if item == self:
                    return_item = self.books
                elif isinstance(item, Books):
                    return_item = item.books[0]
                elif isinstance(item, Book):
                    return_item = item.title
                elif isinstance(item, Field):
                    return_item = item.text
            
            elif axis == self.PreviousSibling:
            
                if isinstance(item, Book):
                    index = item.parent.books.index(item)
                    if index > 0:
                        return_item = item.parent.books[index - 1]
                
                elif isinstance(item, Field):
                    if item.name == u"authors":
                        return_item = item.parent.title
            
            elif axis == self.NextSibling:
            
                if isinstance(item, Book):
                    index = self.books.index(item)
                    if index < len(item.parent.books) - 1:
                        return_item = item.parent.books[index + 1]
                
                elif isinstance(item, Field):
                    if item.name == u"title":
                        return_item = item.parent.authors
        
        if return_item:
            return self._createIndex(return_item)
        else:
            return QXmlNodeModelIndex()


class MessageHandler(QAbstractMessageHandler):

    def handleMessage(self, type, description, identifier, sourceLocation):
    
        self.errorBrowser.setHtml(description)


class MainWindow(QMainWindow):

    def __init__(self):
    
        QMainWindow.__init__(self)
        
        self.books = Books([
            Book(u"Qt for Symbian",
                 u"Frank H. P. Fitzek, Tommi Mikkonen, Tony Torp"),
            Book(u"C++ GUI Programming with Qt 4 (2nd Edition)",
                 u"Jasmin Blanchette and Mark Summerfield"),
            Book(u"Qt 4.6 - GUI-Entwicklung mit C++: Das umfassende Handbuch",
                 u"Jürgen Wolf"),
            Book(u"The Book of Qt 4: The Art of Building Qt Applications",
                 u"Daniel Molkentin"),
            Book(u"Foundations of Qt Development",
                 u"Johan Thelin"),
            Book(u"An Introduction to Design Patterns in C++ with Qt 4",
                 u"Alan Ezust, Paul Ezust"),
            Book(u"Rapid GUI Programming with Python and Qt",
                 u"Mark Summerfield")
            ])
        
        queries = [
            ('Whole document', '$root'),
            ('All books', '$root//book'),
            ('All titles', '$root//title'),
            ('Qt for Symbian authors', '$root//title/text()[string() = "Qt for Symbian"]/../../authors'),
            ('Books with short titles', '$root//title/text()[string-length(string()) < 30]/../..'),
            ('Reformat document',
             '<books>\n'
             '{\n'
             '    for $book in $root//book\n'
             '        return <book title="{$book/title/text()}" authors="{$book/authors/text()}" />\n'
             '}\n'
             '</books>')
            ]
        
        fileMenu = self.menuBar().addMenu(self.tr("&File"))
        fileMenu.addAction(self.tr("E&xit"), self.close, QKeySequence(self.tr("Ctrl+Q")))
        
        insertMenu = self.menuBar().addMenu(self.tr("&Insert"))
        for label, query in queries:
            action = QAction(label, self)
            action.setData(QVariant(query))
            action.triggered.connect(self.insertQuery)
            insertMenu.addAction(action)
        
        queryMenu = self.menuBar().addMenu(self.tr("&Query"))
        queryMenu.addAction(self.tr("&Execute"), self.executeQuery, QKeySequence(self.tr("Ctrl+Return")))
        
        self.inputBrowser = QPlainTextEdit()
        self.inputBrowser.setTabStopWidth(4)
        self.inputBrowser.setReadOnly(True)
        inputText = u"books:\n"
        for book in self.books.books:
            inputText += u"    book:\n        title: %s\n        authors: %s\n" % (
                book.title.text.data, book.authors.text.data)
        self.inputBrowser.setPlainText(inputText)
        
        self.resultBrowser = QTextBrowser()
        self.resultBrowser.setTabStopWidth(4)
        self.resultBrowser.setReadOnly(True)
        self.messageHandler = MessageHandler()
        self.messageHandler.errorBrowser = self.resultBrowser
        
        self.queryEdit = QPlainTextEdit()
        self.queryEdit.setTabStopWidth(4)
        
        splitter = QSplitter(Qt.Vertical)
        splitter.addWidget(self.inputBrowser)
        splitter.addWidget(self.resultBrowser)
        splitter.addWidget(self.queryEdit)
        
        self.setCentralWidget(splitter)
        self.setWindowTitle(self.tr("Books"))
    
    def insertQuery(self):
    
        action = self.sender()
        query = action.data().toString()
        self.queryEdit.setPlainText(query)
        self.resultBrowser.clear()
        self.queryEdit.setFocus(Qt.OtherFocusReason)
    
    def executeQuery(self):
    
        # The query needs other components to use its name pool.
        query = QXmlQuery()
        query.setMessageHandler(self.messageHandler)
        self.bookModel = BookModel(self.books, query.namePool())
        query.bindVariable(u"root", QXmlItem(self.bookModel.root()))
        query_string = u"declare variable $root external;\n" + self.queryEdit.toPlainText()
        query.setQuery(query_string)
        
        if query.isValid():
        
            array = QByteArray()
            buf = QBuffer(array)
            buf.open(QIODevice.WriteOnly)
            
            formatter = QXmlFormatter(query, buf)
            
            if query.evaluateTo(formatter):
                self.resultBrowser.setPlainText(QString.fromUtf8(array))
            
            buf.close()


if __name__ == "__main__":

    app = QApplication(sys.argv)
    
    window = MainWindow()
    window.resize(640, 512)
    window.show()
    
    sys.exit(app.exec_())
